"""
Smartsheet-DataFrame
...

This package contains functions to retrieve Smartsheet
reports and sheets as a Pandas DataFrame

"""

# Standard Imports
from typing import Any
import requests
import logging
import warnings
import time

# 3rd-Party Imports
import pandas as pd

# Local Imports
from .exceptions import AuthenticationError

logger = logging.getLogger(__name__)


def get_report_as_df(token: str = None,
                     report_id: int = None,
                     include_row_id: bool = True,
                     include_parent_id: bool = True,
                     report_obj: Any = None) -> pd.DataFrame:
    """
    Get a Smartsheet report as a Pandas DataFrame

    :param token: Smartsheet Personal Access Token
    :type token: str

    :param report_id: ID of report to retrieve
    :type report_id: int

    :param include_row_id: If True, will append a 'row_id' column to the dataframe
            and populate with row id for each row in sheet
    :type include_row_id: bool

    :param include_parent_id: If True, will append a 'parent_id' column to the
            dataframe and populat with parent ID for each nested row
    :type include_parent_id: bool

    :param report_obj: Smartsheet Python SDK Report object
        Should not be included if token and id_ are provided.
        If both token and id_, and obj are provided, obj will be ignored
    :type report_obj: Any

    :return: Pandas DataFrame with report data
    :rtype: pd.DataFrame
    """

    if not (token or report_obj):
        raise ValueError("One of 'token' or 'report_obj' must be included in parameters")

    if token and not report_id:
        try:
            import smartsheet.models
            if isinstance(token, smartsheet.models.sheet.Sheet):
                raise ValueError("Function must be called with the 'report_obj=' keyword argument")
        except ModuleNotFoundError:
            pass

        raise ValueError("A report_id must be included in the parameters if a token is provided")

    if report_obj and report_id:
        warnings.warn("A 'report_id' has been provided along with a 'report_obj' \n" +
                      "The 'sheet_id' parameter will be ignored")

    if token and report_id:
        return _to_dataframe(_get_from_request(token, report_id, type_="REPORT"), include_row_id, include_parent_id)
    elif report_obj:
        return _to_dataframe(report_obj.to_dict(), include_row_id, include_parent_id)


def get_sheet_as_df(token: str = None,
                    sheet_id: int = None,
                    include_row_id: bool = True,
                    include_parent_id: bool = True,
                    sheet_obj: Any = None) -> pd.DataFrame:
    """
    Get a Smartsheet sheet as a Pandas DataFrame

    :param token: Smartsheet personal authentication token
    :type token: str

    :param sheet_id: Smartsheet source sheet ID to get
    :type sheet_id: int

    :param include_row_id: If True, will append a 'row_id' column to the dataframe
            and populate with row id for each row in sheet
    :type include_row_id: bool

    :param include_parent_id: If True, will append a 'parent_id' column to the
            dataframe and populat with parent ID for each nested row
    :type include_parent_id: bool

    :param sheet_obj: Smartsheet Python SDK sheet object
        Should not be included if token and id_ are provided.
        If both token and id_, and obj are provided, obj will be ignored
    :type sheet_obj: Any

    :return: Pandas DataFrame with sheet data
    :rtype: pd.DataFrame
    """

    if not (token or sheet_obj):
        raise ValueError("One of 'token' or 'sheet_obj' must be included in parameters")

    if token and not sheet_id:
        try:
            import smartsheet.models
            if isinstance(token, smartsheet.models.sheet.Sheet):
                raise ValueError("Function must be called with the 'sheet_obj=' keyword argument")
        except ModuleNotFoundError:
            pass

        raise ValueError("A sheet_id must be included in the parameters if a token is provided")

    if sheet_obj and sheet_id:
        warnings.warn("A 'sheet_id' has been provided along with a 'sheet_obj' \n" +
                      "The 'sheet_id' parameter will be ignored")

    if token and sheet_id:
        return _to_dataframe(_get_from_request(token, sheet_id, type_="SHEET"), include_row_id, include_parent_id)
    elif sheet_obj:
        return _to_dataframe(sheet_obj.to_dict(), include_row_id, include_parent_id)


def get_as_df(type_: str,
              token: str = None,
              id_: int = None,
              obj: Any = None,
              include_row_id: bool = True,
              include_parent_id: bool = True) -> pd.DataFrame:
    """
    Get a Smartsheet report or sheet as a Pandas DataFrame

    :param type_: type of object to get. Must be one of 'report' or 'sheet'
    :type type_: str

    :param token: Smartsheet personal authentication token
    :type token: str

    :param id_: Smartsheet object (report or sheet) ID
    :type id_: int

    :param obj: Smartsheet Python SDK report or sheet object
        Should not be included if token and id_ are provided.
        If both token and id_, and obj are provided, obj will be ignored
    :type obj: Any

    :param include_row_id: If True, will append a 'row_id' column to the dataframe
            and populate with row id for each row in sheet
    :type include_row_id: bool

    :param include_parent_id: If True, will append a 'parent_id' column to the
            dataframe and populate with parent ID for each nested row
    :type include_parent_id: bool

    :return: Pandas DataFrame with object data
    :rtype: pd.DataFrame
    """
    if not (token or obj):
        raise ValueError("One of 'token' or 'obj' must be included in parameters")

    if token and not id_:
        try:
            import smartsheet.models
            if isinstance(token, smartsheet.models.sheet.Sheet):
                raise ValueError("Function must be called with the 'sheet_obj=' keyword argument")
        except ModuleNotFoundError:
            pass

        raise ValueError("A sheet_id must be included in the parameters if a token is provided")

    if obj and id_:
        warnings.warn("An 'id' has been provided along with a 'obj' \n" +
                      "The 'id' parameter will be ignored")

    if token and id_:
        return _to_dataframe(_get_from_request(token, id_, type_), include_row_id, include_parent_id)
    elif obj:
        return _to_dataframe(obj.to_dict(), include_row_id, include_parent_id)


def _get_from_request(token: str, id_: int, type_: str) -> dict:
    if type_.upper() == "SHEET":
        url = f"https://api.smartsheet.com/2.0/sheets/{id_}?include=objectValue&level=1"
        logger.debug("Getting sheet request", extra={'id': id_,
                                                     'url': url,
                                                     'object_type': 'sheet'})
    elif type_.upper() == "REPORT":
        url = f"https://api.smartsheet.com/2.0/reports/{id_}?pageSize=50000"
        logger.debug("Getting report request", extra={'id': id_,
                                                      'url': url,
                                                      'object_Type': 'report'})
    else:
        raise ValueError(f"'type_' parameter must be one of SHEET or REPORT. The current value is {type_.upper()}")

    credentials: dict = {"Authorization": f"Bearer {token}"}
    response = _do_request(url, options=credentials)

    return response.json()


def _to_dataframe(object_dict: dict, include_row_id: bool = True, include_parent_id: bool = True) -> pd.DataFrame:
    columns_list = [column['title'] for column in object_dict['columns']]

    if include_parent_id:
        columns_list.insert(0, "parent_id")
    if include_row_id:
        columns_list.insert(0, "row_id")

    rows_list = []

    # Handle empty sheet condition
    if not object_dict.get('rows', None):
        return pd.DataFrame(columns=columns_list)

    for row in object_dict['rows']:
        cells_list = []
        if include_row_id:
            cells_list.append(int(row['id']))
        if include_parent_id:
            cells_list.append(int(row['parentId'])) if 'parentId' in row else cells_list.append('')

        for cell in row['cells']:
            if 'value' in cell:
                cells_list.append(cell['value'])
            elif 'objectValue' in cell:
                cells_list.append(_handle_object_value(cell['objectValue']))
            else:
                cells_list.append('')
        else:
            rows_list.append(cells_list)

    return pd.DataFrame(rows_list, columns=columns_list)


def _do_request(url: str, options: dict, retries: int = 3) -> requests.Response:
    """
    Do the HTTP request, handling rate limit retrying

    :param url: Smartsheet API URL
    :type url: str

    :param options: API request headers
    :type options: dict

    :param retries: Number of retries
    :type retries: int

    :return: Requests response object
    :rtype: requests.Response
    """

    i = 0
    for i in range(retries):
        try:
            response = requests.get(url, headers=options)
            response_json = response.json()

            if response.status_code != 200:
                if response_json['errorCode'] == 1002 or response_json['errorCode'] == 1003 or \
                        response_json['errorCode'] == 1004:
                    raise AuthenticationError("Could not connect using the supplied auth token \n" +
                                              response.text)
                elif response_json['errorCode'] == 4004:
                    logger.debug(f"Rate limit exceeded. Waiting and trying again... {i}")
                    time.sleep(5 + (i * 5))
                    continue
                else:
                    warnings.warn("An unhandled status_code was returned by the Smartsheet API: \n" +
                                  response.text)
                    return
        except AuthenticationError:
            logger.exception("Smartsheet returned an error status code")
            break
        except Exception:
            logger.exception(f"Not able to retrieve get response. Retrying... {i}")
            time.sleep(5 + (i * 5))
            continue
        break
    else:
        raise Exception(f"Could not retrieve request after retrying {i} times")

    return response


def _handle_object_value(object_value: dict) -> str:
    email_list_string: str = ""
    if object_value['objectType'].upper() == "MULTI_CONTACT":
        email_list_string = ', '.join(obj['email'] for obj in object_value['values'])

    return email_list_string
